#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <goui/libgoui.h>
#ifdef WIN32
#include <windows.h>
#include <goui/libgoui_wingdi.h>
#else
#include <unistd.h>
#include <locale.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <goui/libgoui_x11.h>
#endif
#include <fwm/libfwm.h>
#undef display
#include "OneWinApp.h"

#define MAX_SIZE_X      1000 //2048
#define MAX_SIZE_Y      260 //3200
//#define MAX_WSIZE_X     1536
//#define MAX_WSIZE_Y     1024

#include "DebugFontC.inc"

       goui_buffer * buffer; /* shared with app */
static goui_font * debugfont;
static goui_font * debugfont_small;
static goui_syswin_prefs * syswin_prefs;
       goui_syswin * syswin;
       goui_uxy sw_size;
extern goui_font * get_debugfont(void) { return debugfont; }
extern goui_font * get_debugfont_small(void) { return debugfont_small; }


#ifdef WIN32
extern void debuglog(char const * fmt, ...) {
  char buf[1024];
  va_list arg;
  va_start(arg, fmt);
  _vsnprintf(buf, sizeof(buf), fmt, arg);
  buf[sizeof(buf)-1] = 0;
  va_end(arg);
  MessageBox(syswin?syswin->wnd:NULL, buf, "debug", MB_OK);
}
extern void ErrorBox(char const * fmt, ...) {
  char buf[1024];
  va_list arg;
  va_start(arg, fmt);
  _vsnprintf(buf, sizeof(buf), fmt, arg);
  buf[sizeof(buf)-1] = 0;
  va_end(arg);
  MessageBox(syswin?syswin->wnd:NULL, buf, "ERROR", MB_OK|MB_ICONSTOP);
}
extern void FatalError(char const * fmt, ...) {
  char buf[1024];
  va_list arg;
  va_start(arg, fmt);
  _vsnprintf(buf, sizeof(buf), fmt, arg);
  buf[sizeof(buf)-1] = 0;
  va_end(arg);
  MessageBox(syswin?syswin->wnd:NULL, buf, "FATAL ERROR", MB_OK|MB_ICONSTOP);
  exit(-1);
}
#else
extern void debuglog(char const * fmt, ...) {
  struct timeval tv;
  char buf[1024];
  va_list arg;
  va_start(arg, fmt);

  gettimeofday(&tv, NULL);
  fprintf(stderr, "[%.6f] ", tv.tv_sec+0.000001*tv.tv_usec);

  vsnprintf(buf, sizeof(buf), fmt, arg);
  va_end(arg);
  fprintf(stderr,"DEBUGLOG>>> %s\n", buf);
}
extern void ErrorBox(char const * fmt, ...) {
  char buf[1024];
  va_list arg;
  va_start(arg, fmt);
  vsnprintf(buf, sizeof(buf), fmt, arg);
  va_end(arg);
  fprintf(stderr, "ERROR: %s\n", buf);
}
extern void FatalError(char const * fmt, ...) {
  char buf[1024];
  va_list arg;
  va_start(arg, fmt);
  vsnprintf(buf, sizeof(buf), fmt, arg);
  va_end(arg);
  fprintf(stderr, "FATAL ERROR: %s\n", buf);
  exit(-1);
}
#endif


#define szWNDCLASS "NETMON-GUI"
#define szWNDTITLE "NETMON"
static  int HandleEvent_Low(goui_syswin * sw, goui_event * ev);




extern int SysWinClientLoop(client_idle_func idlefunc) {
	unsigned long count;
	int retval;

	count = 0;
	while(1) {
		switch(goui_event_cycle(NULL, 0)) {
		case 0:
			if(idlefunc && (retval=idlefunc(count++))) return retval;
#ifdef WIN32
			Sleep(10);
#else
			usleep(10000);
#endif
			break;
		case 2:
			return 1;
		}
	}
	return 0;
}


extern goui_font * goui_font_create_from_bitmap_5x10(uchar * bitmap, uchar start, uchar end, uchar mirror) {
        goui_font * f;

        if(!(f = malloc(sizeof(goui_font)))) return NULL;
        memset(f, 0, sizeof(goui_font));
        f->type = 0;
        f->cxs = 5;
        f->cys = 10;
        if(!(f->data = malloc(f->cxs*f->cys*256))) { free(f); return NULL; }
        memset(f->data, 0, f->cxs*f->cys*256);
        {
                goui_usize x, y, i;
                uchar cc, res;
                for(i=start; i<=end; i++) {
                        for(y=0; y<10; y++) {
                                cc = bitmap[(i-start)*16+y];
                                for(x=0; x<5; x++) {
                                        if(!mirror) {
                                                if(cc & (1 << x)) res = 1; else res = 0;
                                        } else {
                                                if(cc & (1 << (7-x))) res = 1; else res = 0;
                                        }
                                        f->data[(i*f->cys+y)*f->cxs+x] = res;
                                }
                        }
                }
        }
        return f;
}




#ifdef WIN32
static int WindowProc(goui_syswin * syswin, UINT message, WPARAM wParam, LPARAM lParam);
int WINAPI WinMain(HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode) {
#else
static int WindowProc(goui_syswin * syswin, XEvent * xev);
int main(int argc, char * * argv) {
	setlocale(LC_CTYPE,""); // needed for XLookupString to properly decode rus layout characters
	app_parse_args(argc, argv);
#endif
	debugfont = goui_font_create_from_bitmap_8x16(font8x16, 32, 127, 0);
	if(!debugfont) { debuglog("Graphics initialization failed"); return 0; }
	debugfont_small = goui_font_create_from_bitmap_5x10(font5x10, 32, 127, 0);
	if(!debugfont_small) { debuglog("Graphics initialization failed"); return 0; }
	app_pre_init();

#ifdef WIN32
	syswin_prefs = goui_wingdi_init_syswin_prefs(NULL, WindowProc, hThisInst, LoadIcon(hThisInst, NULL /*MAKEINTRESOURCE(IDI_ICON1)*/), GOUI_CURSOR_ARROW, (HBRUSH)GetStockObject(BLACK_BRUSH), szWNDCLASS);
#else
	syswin_prefs = goui_x11_init_syswin_prefs(NULL, WindowProc);
#endif
	if(!syswin_prefs) FatalError("Unable to register window class");
	LIBFWM_display = goui_init_default_syswin_ctx()->display;
	LIBFWM_Init("fwm-perfmon", NULL);




	syswin = goui_syswin_create(NULL, syswin_prefs, HandleEvent_Low,
#ifdef WIN32
		szWNDTITLE, NULL, 1, 8,
#else
		szWNDTITLE, NULL, 1, 32,
#endif
//		SYSWIN_DEFAULT, SYSWIN_DEFAULT, SYSWIN_DEFAULT, SYSWIN_DEFAULT,
600, 400, 240, 230,
		MAX_SIZE_X,
		MAX_SIZE_Y,
		SYSWIN_FLAG_MERGE_MOUSEMOVES);
	if(!syswin) { ErrorBox("Unable to create client window"); goui_destroy_syswin_prefs(syswin_prefs); return -1; }
	sw_size.x = syswin->xsize;
	sw_size.y = syswin->ysize;
	buffer = syswin->buffer;


	app_init();
	SysWinClientLoop(&app_idle_func);
	goui_syswin_destroy(syswin);
	goui_destroy_syswin_prefs(syswin_prefs);
	app_post_shutdown();
	return 0;
}




#ifdef WIN32
static int WindowProc(goui_syswin * syswin, UINT message, WPARAM wParam, LPARAM lParam) {
	switch(message){
	case WM_COMMAND:
		break;
	case WM_CLOSE:
		SaveWindowRect();
		return DefWindowProc(syswin->wnd,message,wParam,lParam);
	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(syswin->wnd,message,wParam,lParam);
	}
	return 0;
}
#else
extern int set_widget;
static int WindowProc(goui_syswin * sw, XEvent * xev) { // ret.2=quit, other=ok
	if(!xev) { /* window just created */
		if(set_widget && !LIBFWM_SetWClass(sw->window, WCLASS_WIDGET, "fwm-perfmon")) logprint("SetWClass() failed");
	}
	return 0;
}
#endif

extern void InvalidateSysRectAndWait(goui_ssize x1, goui_ssize y1, goui_ssize x2, goui_ssize y2) {
	if(!syswin) return;
	goui_syswin_invalidate_rect(syswin, x1, y1, x2, y2);
}

extern void InvalidateSysRectAndUpdate(goui_ssize x1, goui_ssize y1, goui_ssize x2, goui_ssize y2) {
	if(!syswin) return;
	goui_syswin_invalidate_rect(syswin, x1, y1, x2, y2);
	goui_syswin_update(syswin);
}

extern void SysWinSetTitle(char const * title) {
	goui_syswin_set_titles(syswin, title, title);
}

static  int HandleEvent_Low(goui_syswin * sw, goui_event * ev) {
	if(ev->what==evWinSizeChanged) {
		sw_size.x = syswin->xsize;
		sw_size.y = syswin->ysize;
	}
	app_handle_event(ev);
	return 0;
}


