#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <goui/libgoui.h>
#include "OneWinApp.h"
#include "util.h"
#include "mon_ifstat.h"
#include "get_memstat.h"
#include "get_cpustat.h"

static void redraw(void);

goui_font * sysfont;

extern goui_buffer * buffer;
int set_widget = 1;
int show_numbers = 1; /* show numeric cpu/mem stats: 0=never; 1=when mouse hover; 2=always */
int if_mode = 0; /* 0=show graph; 1=show interfaces list */

extern void app_parse_args(int argc, char * * argv) {
  int j;
  for(j=1; j<argc; j++) {
    if(!strcmp(argv[j],"no-widget")) set_widget = 0;
    if(!strcmp(argv[j],"widget")) set_widget = 1;
    if(!strcmp(argv[j],"no-numbers")) show_numbers = 0;
    if(!strcmp(argv[j],"mouse-numbers")) show_numbers = 1;
    if(!strcmp(argv[j],"force-numbers")) show_numbers = 2;
  }
}

extern void app_pre_init(void) {
	sysfont = goui_font_load_from_file_8x16(("./cp1251-8x16"), 1);
	if(!sysfont) sysfont = get_debugfont();
}

extern void app_init(void) {
  redraw();
}

extern void app_post_shutdown(void) {
}

static memstat m_stat[3600];
static cpustat c_stat[3600];
static unsigned int stat_pos;
static time_t stat_time;

static bps_t scale_in=1, scale_out=1;
static char ifname[MAX_IFNAME+1] = "ppp0";
static int mouse_zone, iflist_mouse_zone;

static int is_localhost(char const *n) {
  if(n[0]=='l' && n[1]=='o' && (!n[2] || n[2]>='0' && n[2]<='9')) return 1;
  return 0;
}
static void set_if(char const *name) {
  strncpy(ifname, name, sizeof(ifname)-1);
}
static void switch_if(void) {
  uint j;
  if_stats *IF, *IF2;
  for(j=0,IF=ifstats; j<MAX_IF && *IF->name; j++,IF++) if(!strcmp(IF->name,ifname)) { j++; IF++; break; }
  if(j>=MAX_IF || !*IF->name) { j=0; IF=ifstats; }
  if(j>=MAX_IF || !*IF->name) return;
  if(is_localhost(IF->name)) {
    IF2 = IF;
    do {
      j++; IF++;
      if(j>=MAX_IF || !*IF->name) { j=0; IF=ifstats; }
      if(j>=MAX_IF || !*IF->name) return;
    } while(IF!=IF2 && is_localhost(IF->name));
  }
  set_if(IF->name);
}

#define Y_IN 50
#define Y_OU 110
#define Y_MEM 170
#define Y_CPU 228
#define Y_MAX 240
#define Y_H 50

static int get_iflist_mz(goui_event *ev);
static int get_mz(goui_event * ev) {
  if(ev->mousey>0 && ev->mousey<110) return 1;
  else if(ev->mousey>Y_MEM-Y_H && ev->mousey<Y_MEM) return 2;
  else if(ev->mousey>Y_CPU-Y_H && ev->mousey<Y_CPU) return 3;
  else return 0;
}
static int update_mz(goui_event * ev) {
  int mz, imz;
  mz = get_mz(ev);
  imz = get_iflist_mz(ev);
  if(mz!=mouse_zone || imz!=iflist_mouse_zone) {
    mouse_zone = mz;
    iflist_mouse_zone = imz;
    return 1;
  }
  return 0;
}

extern void app_handle_event(goui_event * ev) {
  int need_redraw;
  need_redraw = 0;
  switch(ev->what) {
  case evWinSizeChanged:
    need_redraw = 1;
    break;
  case evMouseLDown:
    need_redraw = update_mz(ev);
    switch(mouse_zone) {
    case 1:
      if(!if_mode) {
        switch_if(); need_redraw = 1;
      } else if(iflist_mouse_zone>=0 && iflist_mouse_zone<MAX_IF && ifstats[iflist_mouse_zone].name[0]) {
        set_if(ifstats[iflist_mouse_zone].name);
        if_mode = 0;
        iflist_mouse_zone = -1;
        need_redraw = 1;
      } else {
        /* switch_if(); need_redraw = 1; */
      }
      break;
    case 2: case 3: show_numbers = (show_numbers+1)%3; need_redraw = 1; break;
    }
    break;
  case evMouseRDown:
    need_redraw = update_mz(ev);
    switch(mouse_zone) {
    case 1: if_mode = !if_mode; update_mz(ev); need_redraw = 1; break;
    }
    break;
  case evMouseMove:
    need_redraw = update_mz(ev);
    break;
  }
  if(need_redraw) redraw();
}


static unsigned int gety(bps_t bps, bps_t sc) {
  double y;
  if(bps<1) return 0;
  y = log10(bps*0.005/sc)*15;
  if(y<1) return 1;
  if(y>Y_H-2) return Y_H-2;
  return (unsigned int)y;
}

static void print_io_num(unsigned int y, ubps_t p) {
  char t[50];
  unsigned int x;
  if(p<1000000) snprintf(t, sizeof(t), "%uk", (uint)(p/1000));
  else if(p<1000000000) snprintf(t, sizeof(t), "%uM", (uint)(p/1000000));
  else snprintf(t, sizeof(t), "%uG", (uint)(p/1000000000));
  x = 14;
  if(strlen(t)>3) x = 8;
  goui_rgb32_text(buffer, x, y-3, t, strlen(t), get_debugfont_small(), 0x888888, 0, 0);
}

static void print_io_num2(uint x, uint y, char const *prefix, ubps_t p) {
  char t[50];
  if(p<1000000) snprintf(t, sizeof(t), "%s%ukB/s", prefix, (uint)(p/1000));
  else if(p<1000000000) snprintf(t, sizeof(t), "%s%uMB/s", prefix, (uint)(p/1000000));
  else snprintf(t, sizeof(t), "%s%uGB/s", prefix, (uint)(p/1000000000));
  goui_rgb32_text(buffer, x, y, t, strlen(t), get_debugfont_small(), 0x888888, 0xFF000000, 0);
}

static void draw_ifstat_bg(void) {
  uint p, y;
  goui_rgb32_text(buffer, 3, Y_IN+1, ifname, strlen(ifname), get_debugfont_small(), 0xFFFFFF, 0, 0);
  goui_rgb32_text(buffer, 3, Y_IN-20, "in", 2, sysfont, 0xFFFFFF, 0, 0);
  goui_rgb32_text(buffer, 3, Y_OU-20, "out", 3, sysfont, 0xFFFFFF, 0, 0);
  for(p=1000; p<1000000; p*=10) {
    y = gety(p,1);
    goui_rgb32_line(buffer, 30, Y_IN-y, sw_size.x, Y_IN-y, 0x333333);
    goui_rgb32_line(buffer, 30, Y_OU-y, sw_size.x, Y_OU-y, 0x333333);
    if(p>1000) print_io_num(Y_IN-y, p*scale_in);
    if(p>1000) print_io_num(Y_OU-y, p*scale_out);
  }
}

static void draw_ifstat_values(uint x, bps_t *v) {
  bps_t rv, sv;
  rv = v?v[0]:-1;
  sv = v?v[1]:-1;
  if(rv>=0 && sv>=0) {
    goui_rgb32_line(buffer, x, Y_IN, x, Y_IN-gety(rv,scale_in), 0x777777);
    goui_rgb32_line(buffer, x, Y_OU, x, Y_OU-gety(sv,scale_out), 0x777777);
  } else {
    goui_rgb32_line(buffer, x, Y_IN, x, Y_IN, 0xEE2222);
    goui_rgb32_line(buffer, x, Y_OU, x, Y_OU, 0xEE2222);
  }
}

static int get_iflist_mz(goui_event *ev) {
  if(if_mode!=1 || ev->mousex<10 || ev->mousey<4 || ev->mousey>102) return -1;
  return (ev->mousey-4)/33 + (ev->mousex-10)/70*3;
}

static void draw_iflist(void) {
  uint x, y, j, p;
  if_stats *IF;
  bps_t *v;
  ull avg_in, avg_out;
  char st[30];
  
  for(j=0,IF=ifstats; j<MAX_IF && *IF->name; j++,IF++) {
    y = (j%3)*33+4;
    x = (j/3)*70+10;
    if(x>=sw_size.x) break;
    if(iflist_mouse_zone>=0 && (uint)iflist_mouse_zone==j) goui_rgb32_fill(buffer, x-4, y+1, x+66, y+34, 0x0000AA);
    else if(!strcmp(ifname,IF->name)) goui_rgb32_fill(buffer, x-4, y+1, x+66, y+34, 0x002844);
    goui_rgb32_text(buffer, x, y, IF->name, strnlen(IF->name, sizeof(IF->name)), sysfont/*get_debugfont_small()*/, 0xFFFFFF, 0xFF000000, (IF->missing_time?GOUI_FONT_STRIKE:0));
    if(IF->missing_time) {
      if(IF->missing_time<100) snprintf(st, sizeof(st), "%u sec", IF->missing_time);
      else if(IF->missing_time<6000) snprintf(st, sizeof(st), "%u min", IF->missing_time/60);
      else if(IF->missing_time<360000) snprintf(st, sizeof(st), "%u hour", IF->missing_time/3600);
      else if(IF->missing_time<3600*24*1000) snprintf(st, sizeof(st), "%u days", IF->missing_time/(3600*24));
      else snprintf(st, sizeof(st), "gone");
      goui_rgb32_text(buffer, x, y+17, st, strlen(st), get_debugfont_small(), 0xFF1111, 0xFF000000, 0);
    } else if(IF->idle_time>60) {
      if(IF->idle_time<100) snprintf(st, sizeof(st), "idle %u sec", IF->idle_time);
      else if(IF->idle_time<6000) snprintf(st, sizeof(st), "idle %u min", IF->idle_time/60);
      else if(IF->idle_time<360000) snprintf(st, sizeof(st), "idle %u hour", IF->idle_time/3600);
      else if(IF->idle_time<3600*24*1000) snprintf(st, sizeof(st), "idle %u days", IF->idle_time/(3600*24));
      else snprintf(st, sizeof(st), "idle");
      goui_rgb32_text(buffer, x, y+17, st, strlen(st), get_debugfont_small(), 0x999999, 0xFF000000, 0);
    } else {
      avg_in = 0; avg_out = 0;
      for(p=(ifstat_pos+3600-10)%3600; p!=ifstat_pos; p=(p+1)%3600) {
        v = IF->bps[p];
        if(v[0]>0) avg_in += v[0];
        if(v[1]>0) avg_out += v[1];
      }
      avg_in/=10; avg_out/=10;
      print_io_num2(x, y+16, "in  ", avg_in);
      print_io_num2(x, y+25, "out ", avg_out);
    }
  }
}

static void prmem(char *st, size_t sz, ull a, ull b) {
  uint A,B;
  if(!b) {
    if(a<1000) { A=a; snprintf(st, sz, "%uK", A); }
    else if(a<10240) { A=(a*10+512)/1024; snprintf(st, sz, "%u.%uM", A/10, A%10); }
    else if(a<1024000) { A=(a+512)/1024; snprintf(st, sz, "%uM", A); }
    else if(a<10485760) { A=(a*10+524288)/1048576; snprintf(st, sz, "%u.%uG", A/10, A%10); }
    else /*if(a<1048576000) */{ A=(a+524288)/1048576; snprintf(st, sz, "%uG", A); }
  } else {
    if(b<1000) { A=a; B=b; snprintf(st, sz, "%u/%uK", A, B); }
    else if(b<10240) { A=(a*10+512)/1024; B=(b*10+512)/1024; snprintf(st, sz, "%u.%u/%u.%uM", A/10, A%10, B/10, B%10); }
    else if(b<1024000) { A=(a+512)/1024; B=(b+512)/1024; snprintf(st, sz, "%u/%uM", A, B); }
    else if(b<10485760) { A=(a*10+524288)/1048576; B=(b*10+524288)/1048576; snprintf(st, sz, "%u.%u/%u.%uG", A/10, A%10, B/10, B%10); }
    else /*if(b<1048576000) */{ A=(a+524288)/1048576; B=(b+524288)/1048576; snprintf(st, sz, "%u/%uG", A, B); }
  }
}

static void draw_mem_bg(uint y) {
  goui_rgb32_text(buffer, 3, y-16, "mem", 3, sysfont, 0xFFFFFF, 0, 0);
  goui_rgb32_text(buffer, 5, y-21, "swap", 4, get_debugfont_small(), 0xFF0000, 0, 0);
  goui_rgb32_text(buffer, 5, y-30, "used", 4, get_debugfont_small(), 0x777777, 0, 0);
  goui_rgb32_text(buffer, 5, y-39, "cache", 5, get_debugfont_small(), 0x3333BB, 0, 0);
  goui_rgb32_text(buffer, 5, y-48, "avail", 5, get_debugfont_small(), 0x33BBBB, 0, 0);
  goui_rgb32_text(buffer, 5, y-57, "free", 4, get_debugfont_small(), 0x22FF22, 0, 0);
}

static ull ull_sub(ull a, ull b) { a -= b; return (0>(long long)a)?0:a; }

static void draw_mem_usage_numbers(uint x, uint y, memstat *m) {
  ull cur, inc;
  char st[20];
  
  if(!m->total) return;
  prmem(st,sizeof(st),cur=m->free,0);
  goui_rgb32_text(buffer, x, y-57, st, strlen(st), get_debugfont_small(), 0x22FF22, 0, 0);
  prmem(st,sizeof(st),inc=ull_sub(m->avail,cur),0); cur+=inc;
  goui_rgb32_text(buffer, x, y-48, st, strlen(st), get_debugfont_small(), 0x33BBBB, 0, 0);
  prmem(st,sizeof(st),inc=ull_sub(m->free+m->temp,cur),0); cur+=inc;
  goui_rgb32_text(buffer, x, y-39, st, strlen(st), get_debugfont_small(), 0x3333BB, 0, 0);
  prmem(st,sizeof(st),ull_sub(m->total,cur),0);
  goui_rgb32_text(buffer, x, y-30, st, strlen(st), get_debugfont_small(), 0x777777, 0, 0);
  prmem(st,sizeof(st),ull_sub(m->swap_total,m->swap_free),m->swap_total);
  goui_rgb32_text(buffer, x, y-21, st, strlen(st), get_debugfont_small(), 0xFF0000, 0, 0);
  prmem(st,sizeof(st),m->total,0);
  goui_rgb32_text(buffer, x, y-12, st, strlen(st), get_debugfont_small(), 0xFFFFFF, 0, 0);
}

static void draw_mem_usage(unsigned int x, unsigned int y, unsigned int h, memstat * m, unsigned int * yold) {
  double mTOT, mAVA, mFREE, mTEMP, sTOT, sFREE;
  unsigned int y1, y2, y3, y4, y5;

  if(!m->total) { goui_rgb32_line(buffer, x, y, x, y, 0xEE2222); *yold = 0; return; }
  mTOT = m->total;
  mTEMP = m->temp;
  mAVA = m->avail;
  mFREE = m->free;
  sTOT = m->swap_total;
  sFREE = m->swap_free;
  
  y1 = (unsigned int)(h*mFREE/mTOT);
  y2 = (unsigned int)(h*mAVA/mTOT);
  y3 = (unsigned int)(h*(mFREE+mTEMP)/mTOT);
  y4 = (unsigned int)(h*mTOT/mTOT);
  y5 = (unsigned int)(h*sFREE/sTOT);
  if(y5>=h) y5=0; else y5=h-y5;
  
  goui_rgb32_line(buffer, x, y-h, x, y-h+y4, 0x777777);
  goui_rgb32_line(buffer, x, y-h, x, y-h+y3, 0x3333BB);
  goui_rgb32_line(buffer, x, y-h, x, y-h+y2, 0x33BBBB);
  goui_rgb32_line(buffer, x, y-h, x, y-h+y1, 0x22FF22);
  if(*yold) goui_rgb32_line(buffer, x+1, y-(*yold-1), x, y-y5, 0xFF0000);
  *yold = y5+1;
}

static void draw_cpu_bg(uint y) {
  goui_rgb32_text(buffer, 3, y-16, "cpu", 3, sysfont, 0xFFFFFF, 0, 0);
  goui_rgb32_text(buffer, 5, y-21, "sys", 3, get_debugfont_small(), 0xBB3333, 0, 0);
  goui_rgb32_text(buffer, 5, y-30, "user", 4, get_debugfont_small(), 0x3377BB, 0, 0);
  goui_rgb32_text(buffer, 5, y-39, "nice", 4, get_debugfont_small(), 0x88BBBB, 0, 0);
  goui_rgb32_text(buffer, 5, y-48, "unkn", 4, get_debugfont_small(), 0x777777, 0, 0);
  goui_rgb32_text(buffer, 5, y-57, "free", 4, get_debugfont_small(), 0x22FF22, 0, 0);
}

static void cpu_pct(char *dst, ull tot, ull val) {
  uint v;
  if(val<tot/10) {
    v = (uint)(2000*val/tot);
    if(v>=200) v = 100; else v = (v+1)/2;
    if(v>=100) strcpy(dst,"10%");
    else snprintf(dst,20,"%u.%u%%", v/10, v%10);
  } else {
    v = (uint)(200*val/tot);
    if(v>=200) v = 100; else if(v>=197) v = 99; else v = (v+1)/2;
    snprintf(dst,20,"%u%%", v);
  }
}

static void draw_cpu_usage_numbers(uint x, uint y, cpustat *prev, cpustat *cur) {
  char st[20];
  ull d_total, d_user, d_nice, d_system, d_idle, d_irq, d_softirq, d_iowait;
  if(!prev->ticks_total || !cur->ticks_total) return;

  d_total   = ull_sub(cur->ticks_total,   prev->ticks_total);
  d_user    = ull_sub(cur->ticks_user,    prev->ticks_user);
  d_nice    = ull_sub(cur->ticks_nice,    prev->ticks_nice);
  d_system  = ull_sub(cur->ticks_system,  prev->ticks_system);
  d_idle    = ull_sub(cur->ticks_idle,    prev->ticks_idle);
  d_iowait  = ull_sub(cur->ticks_iowait,  prev->ticks_iowait);
  d_irq     = ull_sub(cur->ticks_irq,     prev->ticks_irq);
  d_softirq = ull_sub(cur->ticks_softirq, prev->ticks_softirq);
  cpu_pct(st, d_total, d_system+d_irq+d_softirq);
  goui_rgb32_text(buffer, x, y-21, st, strlen(st), get_debugfont_small(), 0xBB3333, 0, 0);
  cpu_pct(st, d_total, d_user);
  goui_rgb32_text(buffer, x, y-30, st, strlen(st), get_debugfont_small(), 0x3377BB, 0, 0);
  cpu_pct(st, d_total, d_nice);
  goui_rgb32_text(buffer, x, y-39, st, strlen(st), get_debugfont_small(), 0x88BBBB, 0, 0);
  cpu_pct(st, d_total, ull_sub(d_total, d_idle+d_system+d_irq+d_softirq+d_user+d_nice));
  goui_rgb32_text(buffer, x, y-48, st, strlen(st), get_debugfont_small(), 0x777777, 0, 0);
  /*
  cpu_pct(st, d_total, d_iowait);
  goui_rgb32_text(buffer, x+40, y-48, st, strlen(st), get_debugfont_small(), 0x777777, 0, 0);
  */
  cpu_pct(st, d_total, d_idle);
  goui_rgb32_text(buffer, x, y-57, st, strlen(st), get_debugfont_small(), 0x22FF22, 0, 0);
}

static void draw_cpu_usage(unsigned int x, unsigned int y, unsigned int h, cpustat * prev, cpustat * cur) {
  ull d_total, d_user, d_nice, d_system, d_idle, d_irq, d_softirq, dd;
  unsigned int cy[4];
  if(!prev->ticks_total || !cur->ticks_total) return;

  d_total   = ull_sub(cur->ticks_total,   prev->ticks_total);
  d_user    = ull_sub(cur->ticks_user,    prev->ticks_user);
  d_nice    = ull_sub(cur->ticks_nice,    prev->ticks_nice);
  d_system  = ull_sub(cur->ticks_system,  prev->ticks_system);
  d_idle    = ull_sub(cur->ticks_idle,    prev->ticks_idle);
  d_irq     = ull_sub(cur->ticks_irq,     prev->ticks_irq);
  d_softirq = ull_sub(cur->ticks_softirq, prev->ticks_softirq);
  dd = d_system+d_irq+d_softirq;
  cy[0] = (unsigned int)(h*dd/d_total); if(cy[0]>h) cy[0]=h;
  dd += d_user;
  cy[1] = (unsigned int)(h*dd/d_total); if(cy[1]>h) cy[1]=h;
  dd += d_nice;
  cy[2] = (unsigned int)(h*dd/d_total); if(cy[2]>h) cy[2]=h;
  dd = d_total - d_idle;
  cy[3] = (unsigned int)(h*dd/d_total); if(cy[3]>h) cy[3]=h;
  goui_rgb32_line(buffer, x, y, x, y-h, 0x22FF22);
  goui_rgb32_line(buffer, x, y, x, y-cy[3], 0x777777);
  goui_rgb32_line(buffer, x, y, x, y-cy[2], 0x88BBBB);
  goui_rgb32_line(buffer, x, y, x, y-cy[1], 0x3377BB);
  goui_rgb32_line(buffer, x, y, x, y-cy[0], 0xBB3333);
}

static void redraw(void) {
  unsigned int x, p, ppp, y, yold, j;
  if_stats *IF;
  bps_t *v;
  bps_t max_in, max_out;

  if(sw_size.x<1 || sw_size.y<1) return;
//  goui_rgb32_fill(buffer, 0, 0, buffer->size.x, buffer->size.y, 0);
  goui_buffer_empty(buffer, 0);
  
  if(if_mode==0) {
    max_in = max_out = 0;
    IF = get_if_by_name(ifname,0);
    if(IF) for(j=1; j<=3590 && sw_size.x-j>=30; j++) {
      v = IF->bps[(ifstat_pos+3600-j)%3600];
      if(max_in<v[0]) max_in=v[0];
      if(max_out<v[1]) max_out=v[1];
    }
    /* scale=1 means up to ~100k graph, scale=1000000 means 100G */
    scale_in = 1;
    scale_out = 1;
    while(max_in>200000 && scale_in<1000000) { max_in/=10; scale_in*=10; }
    while(max_out>200000 && scale_out<1000000) { max_out/=10; scale_out*=10; }
    draw_ifstat_bg();
  } else {
    draw_iflist();
  }
  
  draw_mem_bg(Y_MEM);
  draw_cpu_bg(Y_CPU);
  
  yold = 0;
  for(j=1; j<=3590; j++) {
    x = sw_size.x-j;
    if(x<30) break;
    if(if_mode==0) {
      if(!(j%60)) goui_rgb32_line(buffer, x, 0, x, Y_MAX-1, 0x333366);
      draw_ifstat_values(x, IF?IF->bps[(ifstat_pos+3600-j)%3600]:NULL);
    } else {
      if(!(j%60)) goui_rgb32_line(buffer, x, Y_OU, x, Y_MAX-1, 0x333366);
    }
    p = (stat_pos+3600-j)%3600;
    ppp = (stat_pos+3599-j)%3600;
    draw_mem_usage(x, Y_MEM, Y_H, m_stat+p, &yold);
    draw_cpu_usage(x, Y_CPU, Y_H, c_stat+ppp, c_stat+p);
  }
  
  if(show_numbers>=2 || show_numbers && mouse_zone==2) {
    p = (stat_pos+3599)%3600;
    draw_mem_usage_numbers(31, Y_MEM, m_stat+p);
  }
  if(show_numbers>=2 || show_numbers && mouse_zone==3) {
    p = (stat_pos+3599)%3600;
    for(j=3589; j<3598; j++) {
      ppp = (stat_pos+j)%3600;
      if(c_stat[ppp].ticks_total) break;
    }
    draw_cpu_usage_numbers(31, Y_CPU, c_stat+ppp, c_stat+p);
  }
  
  InvalidateSysRectAndUpdate(0,0,sw_size.x,sw_size.y);
}

static int update_all(void) {
  time_t now, diff;
  unsigned int prevpos;
  
  time(&now);
  diff = now - stat_time;
  if(!diff) return 0;
  update_ifstats(diff);
  if(diff==1) {
    stat_pos = (stat_pos+1)%3600;
    stat_time++;
  } else if(diff<3600) {
    while(diff--) {
      m_stat[stat_pos] = (memstat) { };
      c_stat[stat_pos] = (cpustat) { };
      stat_pos = (stat_pos+1)%3600;
      stat_time++;
    }
  } else {
    for(stat_pos=3600; stat_pos; ) {
      stat_pos--; 
      m_stat[stat_pos] = (memstat) { };
      c_stat[stat_pos] = (cpustat) { };
    }
    stat_time = now;
  }
  if(stat_pos) prevpos=stat_pos-1; else prevpos=3599;
  if(get_memstat(m_stat+prevpos)<0) m_stat[prevpos] = (memstat) { };
  if(get_cpustat(c_stat+prevpos)<0) c_stat[prevpos] = (cpustat) { };
  return 1;
}

extern int  app_idle_func(unsigned long count) {
  if(update_all()) redraw();
  return 0;
}
