#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "mon_ifstat.h"

#define stat_pos ifstat_pos
#define MAX_SEC MAX_IFSEC
if_stats ifstats[MAX_IF];
uint stat_pos; /* ifstats[].bps[stat_pos] is a place immediately after last stored diff */

static void remove_missing_interfaces(void) {
  uint j, k;
  for(k=j=0; j<MAX_IF && ifstats[j].name[0]; j++) {
    if(ifstats[j].missing_time<MAX_SEC) {
      if(k!=j) bcopy(ifstats+j, ifstats+k, sizeof(*ifstats));
      k++;
    }
  }
  while(k<j) {
    ifstats[k].name[0] = 0;
    k++;
  }
}

extern if_stats * get_if_by_name(char const *ifname, int may_create) {
  uint j;
  if_stats *i;
  for(i=ifstats,j=0; j<MAX_IF && i->name[0]; i++,j++) if(!strncmp(i->name,ifname,MAX_IFNAME)) return i;
  if(may_create && j<MAX_IF) {
    strncpy(i->name,ifname,MAX_IFNAME);
    i->idle_time = 0;
    i->missing_time = 0;
    i->avg_bp10m = 0;
    i->last_valid = 0;
    i->next_valid = 0;
    memset(i->bps, -1, sizeof(i->bps));
    return i;
  }
  return NULL;
}

static void scan_ifstats(void) {
  FILE * fp;
  char buf[4000];
  char *p, *ifname;
  ull brecv, bsent;
  if_stats *i;
  if(!(fp = fopen("/proc/net/dev", "rb"))) return;
  while(fgets(buf, sizeof(buf), fp)) {
    p = buf;
    while(*p==' ') p++;
    ifname = p;
    if(!(p = strchr(ifname,':'))) continue;
    *(p++) = 0;
    while(*p==' ') p++;
    if(parse_ull(&p, &brecv)<0) continue;
    if(*p!=' ') continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(!(p=strchr(p,' '))) continue; while(*p==' ') p++;
    if(parse_ull(&p, &bsent)<0) continue;
    if(*p!=' ') continue;
    if(!(i = get_if_by_name(ifname,1))) continue;
    i->next_valid = 1;
    i->next_recv = brecv;
    i->next_sent = bsent;
  }
  fclose(fp);
}

static void safe_add(uint *v, uint add) {
  uint t;
  t = (*v) + add;
  if(t>=add) *v = t;
  else *v = (uint)-1;
}

extern void update_ifstats(uint timediff) {
  uint j, k;
  if_stats *i;
  bps_t brecv, bsent;
  ubps_t bps;
  if(!timediff) return;
  remove_missing_interfaces();
  scan_ifstats();
  if(timediff==1) {
    for(i=ifstats,j=0; j<MAX_IF && i->name[0]; i++,j++) {
      if(i->last_valid && i->next_valid) {
        brecv = i->next_recv-i->last_recv;
        bsent = i->next_sent-i->last_sent;
      } else {
        brecv = -1;
        bsent = -1;
      }
      if(i->next_valid) i->missing_time = 0; else safe_add(&i->missing_time, 1);
      if(brecv>0 || bsent>0) i->idle_time = 0; else safe_add(&i->idle_time, 1);
      bps = ((brecv>0)?brecv:0)+((bsent>0)?bsent:0);
      if(i->avg_bp10m>600) i->avg_bp10m -= (i->avg_bp10m+300)/600; else if(i->avg_bp10m) i->avg_bp10m--;
      i->avg_bp10m += bps;
      i->bps[stat_pos][0] = brecv;
      i->bps[stat_pos][1] = bsent;
      i->last_recv = i->next_recv;
      i->last_sent = i->next_sent;
      i->last_valid = i->next_valid;
      i->next_valid = 0;
    }
    stat_pos = (stat_pos+1)%MAX_SEC;
    return;
  }
  for(i=ifstats,j=0; j<MAX_IF && i->name[0]; i++,j++) {
    if(i->next_valid) i->missing_time = 0; else safe_add(&i->missing_time, timediff);
    safe_add(&i->idle_time, timediff);
    if(timediff<3600) {
      for(k=0; k<timediff; k++) {
        if(i->avg_bp10m>600) i->avg_bp10m -= (i->avg_bp10m+300)/600; else if(i->avg_bp10m) i->avg_bp10m--;
        i->bps[(stat_pos+k)%MAX_SEC][0] = -1;
        i->bps[(stat_pos+k)%MAX_SEC][1] = -1;
      }
    } else {
      i->avg_bp10m = 0;
      memset(i->bps, -1, sizeof(i->bps));
    }
    i->last_recv = i->next_recv;
    i->last_sent = i->next_sent;
    i->last_valid = i->next_valid;
    i->next_valid = 0;
  }
  stat_pos = (stat_pos+timediff)%MAX_SEC;
}

